/*
** anno.c -- annotate messages
**
** This code is Copyright (c) 2002, by the authors of nmh.  See the
** COPYRIGHT file in the root directory of the nmh distribution for
** complete copyright information.
**
** Three new options have been added: delete, list, and number.
** Message header fields are used by the new MIME attachment code in
** the send command.  Adding features to generalize the anno command
** seemed to be a better approach than the creation of a new command
** whose features would overlap with those of the anno command.
**
** The -delete option deletes header elements that match the -component
** field name.  If -delete is used without the -text option, the first
** header field whose field name matches the component name is deleted.
** If the -delete is used with the -text option, and the -text argument
** begins with a /, the first header field whose field name matches the
** component name and whose field body matches the text is deleted.  If
** the -text argument does not begin with a /, then the text is assumed
** to be the last component of a path name, and the first header field
** whose field name matches the component name and a field body whose
** last path name component matches the text is deleted.  If the -delete
** option is used with the new -number option described below, the nth
** header field whose field name matches the component name is deleted.
** No header fields are deleted if none of the above conditions are met.
**
** The -list option outputs the field bodies from each header field whose
** field name matches the component name, one per line.  If no -text
** option is specified, only the last path name component of each field
** body is output.  The entire field body is output if the -text option
** is used; the contents of the -text argument are ignored.  If the -list
** option is used in conjuction with the new -number option described
** below, each line is numbered starting with 1.  A tab separates the
** number from the field body.
**
** The -number option works with both the -delete and -list options as
** described above.  The -number option takes an optional argument.  A
** value of 1 is assumed if this argument is absent.
*/

#include <h/mh.h>
#include <h/utils.h>
#include <h/tws.h>
#include <fcntl.h>
#include <errno.h>
#include <utime.h>


static struct swit switches[] = {
#define COMPSW 0
	{ "component field", 0 },
#define DATESW 1
	{ "date", 0 },
#define NDATESW 2
	{ "nodate", 0 },
#define TEXTSW 3
	{ "text body", 0 },
#define VERSIONSW 4
	{ "version", 0 },
#define HELPSW 5
	{ "help", 0 },
#define LISTSW 6
	{ "list", 1 },
#define DELETESW 7
	{ "delete", 2 },
#define NUMBERSW 8
	{ "number", 2 },
#define APPENDSW 9
	{ "append", 1 },
#define PRESERVESW 10
	{ "preserve", 1 },
#define NOPRESERVESW 11
	{ "nopreserve", 3 },
	{ NULL, 0 }
};

/*
** static prototypes
*/
static void make_comp(unsigned char **);
static int annosbr(int, char *, char *, char *, int, int, int);
static int annotate(char *, char *, char *, int, int, int, int);
static void annolist(char *, char *, char *, int);


int
main(int argc, char **argv)
{
	int datesw = 1;
	int preserve = 0;
	int msgnum;
	char *cp, *maildir;
	unsigned char *comp = NULL;
	char *text = NULL, *folder = NULL, buf[BUFSIZ];
	char *file = NULL;
	char **argp, **arguments;
	struct msgs_array msgs = { 0, 0, NULL };
	struct msgs *mp;
	int append = 0;  /* append annotations instead of default prepend */
	int delete = -2;  /* delete header element if set */
	int list = 0;  /* list header elements if set */
	int number = 0; /* delete specific number of like elements if set */
	int havemsgs = 0;

#ifdef LOCALE
	setlocale(LC_ALL, "");
#endif
	invo_name = mhbasename(argv[0]);

	/* read user profile/context */
	context_read();

	arguments = getarguments(invo_name, argc, argv, 1);
	argp = arguments;

	while ((cp = *argp++)) {
		if (*cp == '-') {
			switch (smatch(++cp, switches)) {
			case AMBIGSW:
				ambigsw(cp, switches);
				done(1);
			case UNKWNSW:
				adios(NULL, "-%s unknown", cp);

			case HELPSW:
				snprintf(buf, sizeof(buf),
					"%s [+folder] [msgs] [switches]",
					invo_name);
				print_help(buf, switches, 1);
				done(1);
			case VERSIONSW:
				print_version(invo_name);
				done(1);

			case COMPSW:
				if (comp)
					adios(NULL, "only one component at a time!");
				if (!(comp = *argp++) || *comp == '-')
					adios(NULL, "missing argument to %s",
							argp[-2]);
				continue;

			case DATESW:
				datesw++;
				continue;
			case NDATESW:
				datesw = 0;
				continue;

			case TEXTSW:
				if (text)
					adios(NULL, "only one body at a time!");
				if (!(text = *argp++) || *text == '-')
					adios(NULL, "missing argument to %s",
							argp[-2]);
				continue;

			case DELETESW:  /* delete annotations */
				delete = 0;
				continue;

			case LISTSW:  /* produce a listing */
				list = 1;
				continue;

			case NUMBERSW: /* number listing or delete by number */
				if (number != 0)
					adios(NULL, "only one number at a time!");

				if (argp-arguments == argc-1 || **argp == '-')
					number = 1;

				else {
					if (strcmp(*argp, "all") == 0)
						number = -1;
					else if (!(number = atoi(*argp)))
						/* FIXME: fails for
						** `-list -number l:10'
						** but okay if we add `all'.
						*/
						adios(NULL, "missing argument to %s", argp[-1]);
					argp++;
				}

				delete = number;
				continue;

			case APPENDSW:
				append = 1;
				continue;

			case PRESERVESW:
				preserve = 1;
				continue;

			case NOPRESERVESW:
				preserve = 0;
				continue;
			}
		}
		if (*cp == '+' || *cp == '@') {
			if (folder)
				adios(NULL, "only one folder at a time!");
			else
				folder = getcpy(expandfol(cp));
		} else if (*cp == '/' || *cp == '.') {
			if (file)
				adios(NULL, "only one file at a time!");
			file = cp;
		} else {
			app_msgarg(&msgs, cp);
			havemsgs = 1;
		}
	}

	if (file && (folder || havemsgs)) {
		adios(NULL, "Don't intermix files and messages.");
	}

	make_comp(&comp);

	if (file) {
		if (list)
			annolist(file, comp, text, number);
		else
			annotate(file, comp, text,
					datesw, delete, append, preserve);
		done(0);
	}

	if (!msgs.size)
		app_msgarg(&msgs, seq_cur);
	if (!folder)
		folder = getcurfol();
	maildir = toabsdir(folder);

	if (chdir(maildir) == NOTOK)
		adios(maildir, "unable to change directory to");

	/* read folder and create message structure */
	if (!(mp = folder_read(folder)))
		adios(NULL, "unable to read folder %s", folder);

	/* check for empty folder */
	if (mp->nummsg == 0)
		adios(NULL, "no messages in %s", folder);

	/* parse all the message ranges/sequences and set SELECTED */
	for (msgnum = 0; msgnum < msgs.size; msgnum++)
		if (!m_convert(mp, msgs.msgs[msgnum]))
			done(1);

	/* annotate all the SELECTED messages */
	for (msgnum = mp->lowsel; msgnum <= mp->hghsel; msgnum++) {
		if (is_selected(mp, msgnum)) {
			if (list)
				annolist(m_name(msgnum), comp, text, number);
			else
				annotate(m_name(msgnum), comp, text, datesw,
						delete, append, preserve);
		}
	}

	context_replace(curfolder, folder);
	seq_setcur(mp, mp->lowsel);
	seq_save(mp);
	folder_free(mp);
	context_save();
	done(0);
	return 1;
}

static void
make_comp(unsigned char **ap)
{
	register unsigned char *cp;
	char buffer[BUFSIZ];

	if (*ap == NULL) {
		printf("Enter component name: ");
		fflush(stdout);

		if (fgets(buffer, sizeof buffer, stdin) == NULL)
			done(1);
		*ap = trimcpy(buffer);
	}

	if ((cp = *ap + strlen(*ap) - 1) > *ap && *cp == ':')
		*cp = 0;
	if (strlen(*ap) == 0)
		adios(NULL, "null component name");
	if (**ap == '-')
		adios(NULL, "invalid component name %s", *ap);
	if (strlen(*ap) >= NAMESZ)
		adios(NULL, "too large component name %s", *ap);

	for (cp = *ap; *cp; cp++)
		if (!isalnum(*cp) && *cp != '-')
			adios(NULL, "invalid component name %s", *ap);
}


static int
annotate(char *file, char *comp, char *text, int datesw,
	int delete, int append, int preserve)
{
	int i, fd;
	struct utimbuf b;
	struct stat s;

	/* open and lock the file to be annotated */
	if ((fd = lkopen(file, O_RDWR, 0)) == NOTOK) {
		switch (errno) {
		case ENOENT:
			break;
		default:
			admonish(file, "unable to lock and open");
			break;
		}
		return 1;
	}

	if (stat(file, &s) == -1) {
		advise("can't get access and modification times for %s", file);
		preserve = 0;
	}

	b.actime = s.st_atime;
	b.modtime = s.st_mtime;

	i = annosbr(fd, file, comp, text, datesw, delete, append);

	if (preserve && utime(file, &b) == -1) {
		advise("can't set access and modification times for %s", file);
	}
	lkclose(fd, file);
	return i;
}

/*
**  Produce a listing of all header fields (annotations) whose field
**  name matches comp.  Number the listing if number is set.  Treat the
**  field bodies as path names and just output the last component unless
**  text is non-NULL.  We don't care what text is set to.
*/
static void
annolist(char *file, char *comp, char *text, int number)
{
	int c;
	int count;  /* header field (annotation) counter */
	char *cp;
	char *field;
	int field_size;
	FILE *fp;
	int length;
	int n;  /* number of bytes written */
	char *sp;

	if ((fp = fopen(file, "r")) == NULL) {
		adios(file, "unable to open");
	}

	/* We'll grow this buffer as needed. */
	field = (char *)mh_xmalloc(field_size = 256);

	length = strlen(comp); /* Convenience copy. */
	count = 0;

	do {
		/*
		** Get a line from the input file, growing the field buffer
		** as needed.  We do this so that we can fit an entire line
		** in the buffer making it easy to do a string comparison
		** on both the field name and the field body which might be
		** a long path name.
		*/
		for (n = 0, cp = field; (c = getc(fp)) != EOF; *cp++ = c) {
			if (c == '\n' && (c = getc(fp)) != ' ' && c != '\t') {
				ungetc(c, fp);
				c = '\n';
				break;
			}
			if (++n >= field_size - 1) {
				field = (char *)mh_xrealloc(field,
						field_size += 256);
				cp = field + n - 1;
			}
		}
		*cp = '\0';

		if (strncasecmp(field, comp, length)==0 &&
				field[length] == ':') {
			for (cp = field + length + 1;
					*cp == ' ' || *cp == '\t'; cp++) {
				continue;
			}
			if (number) {
				printf("%d\t", ++count);
			}
			if (!text && (sp = strrchr(cp, '/'))) {
				cp = sp + 1;
			}
			printf("%s\n", cp);
		}

	} while (*field && *field != '-');

	free(field);
	fclose(fp);

	return;
}

static int
annosbr(int fd, char *file, char *comp, char *text, int datesw, int delete,
		int append)
{
	int mode, tmpfd;
	char *cp, *sp;
	char tmpfil[BUFSIZ];
	struct stat st;
	FILE *tmp;
	int c;
	int count;  /* header field (annotation) counter */
	char *field = NULL;
	int field_size = 0;
	FILE *fp = NULL;
	int length;
	int n;  /* number of bytes written */

	mode = fstat(fd, &st) != NOTOK ? (int)(st.st_mode & 0777) : m_gmprot();

	strncpy(tmpfil, m_mktemp2(file, "annotate", NULL, &tmp),
			sizeof(tmpfil));
	chmod(tmpfil, mode);

	/*
	** We're going to need to copy some of the message file to the
	** temporary file while examining the contents.  Convert the
	** message file descriptor to a file pointer since it's a lot
	** easier and more efficient to use stdio for this.  Also allocate
	** a buffer to hold the header components as they're read in.
	** This buffer is grown as needed later.
	*/
	if (delete >= -1 || append != 0) {
		if ((fp = fdopen(fd, "r")) == NULL) {
			adios(NULL, "unable to fdopen file.");
		}
		field = (char *)mh_xmalloc(field_size = 256);
	}

	/*
	** We're trying to delete a header field (annotation) if the
	** delete flag is greater -2.  A value greater than zero
	** means that we're deleting the nth header field that matches
	** the field (component) name.  A value of zero means that
	** we're deleting the first field in which both the field name
	** matches the component name and the field body matches the text.
	** The text is matched in its entirety if it begins with a slash;
	** otherwise the text is matched against whatever portion of the
	** field body follows the last slash.  This allows matching of
	** both absolute and relative path names.  This is because this
	** functionality was added to support attachments.  It might be
	** worth having a separate flag to indicate path name matching
	** to make it more general.  A value of -1 means to delete all
	** matching fields.
	*/
	if (delete >= -1) {
		length = strlen(comp);  /* convenience copy */
		count = 0; /* Only used if we're deleting by number. */

		/*
		**  Copy lines from the input file to the temporary file
		**  until we either find the one that we're looking
		**  for (which we don't copy) or we reach the end of
		**  the headers.  Both a blank line and a line beginning
		**  with a - terminate the headers so that we can handle
		**  both drafts and RFC-2822 format messages.
		*/
		do {
			/*
			** Get a line from the input file, growing the
			** field buffer as needed.  We do this so that
			** we can fit an entire line in the buffer making
			** it easy to do a string comparison on both the
			** field name and the field body which might be
			** a long path name.
			*/
			for (n=0, cp=field; (c=getc(fp)) != EOF; *cp++ = c) {
				if (c == '\n' && (c = getc(fp)) != ' ' &&
						c != '\t') {
					ungetc(c, fp);
					c = '\n';
					break;
				}

				if (++n >= field_size - 1) {
					field = (char *) mh_xrealloc(field,
							field_size *= 2);
					cp = field + n - 1;
				}
			}
			*cp = '\0';

			/*
			** Check for a match on the field name.  We delete
			** the line by not copying it to the temporary
			** file if
			**
			**  o  The delete flag is 0, meaning that we're
			**     going to delete the first matching
			**     field, and the text is NULL meaning that
			**     we don't care about the field body.
			**
			**  o  The delete flag is 0, meaning that we're
			**     going to delete the first matching
			**     field, and the text begins with a / meaning
			**     that we're looking for a full path name,
			**     and the text matches the field body.
			**
			**  o  The delete flag is 0, meaning that we're
			**     going to delete the first matching
			**     field, the text does not begin with a /
			**     meaning that we're looking for the last
			**     path name component, and the last path
			**     name component matches the text.
			**
			**  o  The delete flag is positive meaning that
			**     we're going to delete the nth field
			**     with a matching field name, and this is
			**     the nth matching field name.
			**
			**  o  The delete flag is -1 meaning that we're
			**     going to delete all fields with a
			**     matching field name.
			*/
			if (strncasecmp(field, comp, length)==0 &&
					field[length] == ':') {
				if (!delete) {
					if (!text) {
						break;
					}
					for (cp=field+length+1;
							*cp==' ' || *cp=='\t';
							cp++) {
						continue;
					}
					if (*text == '/' &&
							strcmp(cp, text)==0) {
						break;
					} else {
						if ((sp = strrchr(cp, '/'))) {
							cp = sp + 1;
						}
						if (strcmp(cp, text)==0) {
							break;
						}
					}
				} else if (delete == -1) {
					continue;
				} else if (++count == delete) {
					break;
				}
			}

			/*
			** This line wasn't a match so copy it to the
			** temporary file.
			*/
			if ((n = fputs(field, tmp)) == EOF ||
					(c=='\n' && fputc('\n', tmp)==EOF)) {
				adios(NULL, "unable to write temporary file.");
			}
		} while (*field && *field != '-');
		free(field);

	} else {
		/*
		**  Find the end of the headers before adding the
		**  annotations if we're appending instead of the default
		**  prepending.  A special check for no headers is needed
		**  if appending.
		*/
		if (append) {
			/*
			** Copy lines from the input file to the temporary
			** file until we reach the end of the headers.
			*/
			if ((c = getc(fp)) == '\n') {
				rewind(fp);
			} else {
				putc(c, tmp);
				while ((c = getc(fp)) != EOF) {
					putc(c, tmp);
					if (c == '\n') {
						ungetc(c = getc(fp), fp);
						if (c == '\n' || c == '-') {
							break;
						}
					}
				}
			}
		}

		if (datesw) {
			fprintf(tmp, "%s: %s\n", comp, dtimenow(0));
		}
		if ((cp = text)) {
			do {
				while (*cp == ' ' || *cp == '\t') {
					cp++;
				}
				sp = cp;
				while (*cp && *cp++ != '\n') {
					continue;
				}
				if (cp - sp) {
					fprintf(tmp, "%s: %*.*s", comp,
						(int)(cp - sp),
						(int)(cp - sp), sp);
				}
			} while (*cp);
			if (cp[-1] != '\n' && cp != text) {
				putc('\n', tmp);
			}
		}
	}
	fflush(tmp);

	/*
	** We've been messing with the input file position.  Move the
	** input file descriptor to the current place in the file
	** because the stock data copying routine uses the descriptor,
	** not the pointer.
	*/
	if (append || delete >= -1) {
		if (lseek(fd, (off_t)ftell(fp), SEEK_SET) == (off_t)-1) {
			adios(NULL, "can't seek.");
		}
	}

	cpydata(fd, fileno(tmp), file, tmpfil);
	fclose(tmp);

	if ((tmpfd = open(tmpfil, O_RDONLY)) == NOTOK) {
		adios(tmpfil, "unable to open for re-reading");
	}
	lseek(fd, (off_t) 0, SEEK_SET);

	/*
	**  We're making the file shorter if we're deleting a header field
	**  so the file has to be truncated or it will contain garbage.
	*/
	if (delete >= -1 && ftruncate(fd, 0) == -1) {
		adios(tmpfil, "unable to truncate.");
	}
	cpydata(tmpfd, fd, tmpfil, file);
	close(tmpfd);
	unlink(tmpfil);

	/*
	** Close the delete file so that we don't run out of file pointers if
	** we're doing piles of files.  Note that this will make the close() in
	** lkclose() fail, but that failure is ignored so it's not a problem.
	*/
	if (delete >= -1) {
		fclose(fp);
	}
	return 0;
}
